==> mutual_credit-5.2.x-dev.tar.gz: mutual_credit/src/Entity/Wallet.php <==
<?php

namespace Drupal\mcapi\Entity;

use Drupal\mcapi\Entity\Currency;
use Drupal\mcapi\Entity\CurrencyInterface;
use Drupal\user\EntityOwnerTrait;
use Drupal\user\EntityOwnerInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\StringTranslation\StringTranslationTrait;

/**
 * Defines the wallet entity.
 *
 * The metaphor is not perfect because this wallet can hold a negaive balance.
 * A more accurate term, 'account', has another meaning in Drupal.
 *
 * This Entity knows all about the special intertrading wallets though none are
 * installed by default.
 *
 * @ContentEntityType(
 *   id = "mcapi_wallet",
 *   label = @Translation("Wallet"),
 *   module = "mcapi",
 *   handlers = {
 *     "view_builder" = "Drupal\mcapi\Entity\ViewBuilder\WalletViewBuilder",
 *     "storage" = "Drupal\mcapi\Entity\Storage\WalletStorage",
 *     "storage_schema" = "Drupal\mcapi\Entity\Storage\WalletStorageSchema",
 *     "list_builder" = "Drupal\mcapi\Entity\ListBuilder\WalletListBuilder",
 *     "access" = "Drupal\mcapi\Entity\Access\WalletAccessControlHandler",
 *     "form" = {
 *       "default" = "Drupal\mcapi\Form\WalletForm",
 *       "create" = "Drupal\mcapi\Form\WalletAddForm",
 *       "delete" = "Drupal\Core\Entity\ContentEntityDeleteForm",
 *     },
 *     "views_data" = "Drupal\mcapi\Entity\Views\WalletViewsData",
 *     "route_provider" = {
 *       "html" = "Drupal\mcapi\Entity\RouteProvider\WalletRouteProvider",
 *     },
 *   },
 *   admin_permission = "configure mcapi",
 *   base_table = "mcapi_wallet",
 *   entity_keys = {
 *     "label" = "name",
 *     "id" = "wid",
 *     "uuid" = "uuid",
 *     "owner" = "owner"
 *   },
 *   translatable = FALSE,
 *   links = {
 *     "canonical" = "/wallet/{mcapi_wallet}",
 *     "log" = "/wallet/{mcapi_wallet}/log",
 *     "inex" = "/wallet/{mcapi_wallet}/inex",
 *     "delete-form" = "/wallet/{mcapi_wallet}/delete",
 *     "edit-form" = "/wallet/{mcapi_wallet}/edit",
 *     "collection" = "/admin/accounting/wallets"
 *   },
 *   field_ui_base_route = "entity.mcapi_wallet.collection"
 * )
 */
class Wallet extends ContentEntityBase implements WalletInterface, EntityOwnerInterface {

  use MessengerTrait;
  use StringTranslationTrait;
  use EntityOwnerTrait;

  private $stats = NULL;

  const STATS = [
    'balance' => 0,
    'volume' => 0,
    'gross_in' => 0,
    'gross_out' => 0,
    'trades' => 0,
    'partners' => 0
  ];

  /**
   * {@inheritDoc}
   */
  public static function precreate($wallet, array &$values) {
    if (!isset($values['owner'])) {
      $values['owner'] = \Drupal\user\Entity\User::load(1);
    }
    // NB owner is not set when generating a wallet for the field edit page.
    if (!isset($values['name'])) {
      // N.B. If this wallet is autocreated then the owner isn't saved yet
      $values['name'] = $values['owner']->getAccountName();
    }
  }

  /**
   * {@inheritDoc}
   */
  public function label() {
    if (\Drupal::currentUser()->isAnonymous()) {
      return $this->t('Wallet @num', ['@num' => $this->id()]);
    }
    else {
      return $this->name->value;
    }
  }

  /**
   * {@inheritDoc}
   */
  public function doSave($id, EntityInterface $entity) {
    // Check the name length.
    if (strlen($this->name->value) > 128) {
      $this->name->value = substr($this->name->value, 0, 128);
      $this->messenger()->addWarning(
        $this->t('Wallet name was truncated to 128 characters: @name', ['@name' => $this->name->value]),
        'warning'
      );
    }
  }

  /**
   * {@inheritDoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    $fields = parent::baseFieldDefinitions($entity_type)
      + static::ownerBaseFieldDefinitions($entity_type);
    $fields['owner']->setDisplayConfigurable('form', TRUE)
      ->setDisplayConfigurable('view', TRUE)
      ->setLabel(t('Wallet owner', [], ['context' => 'The user who owns the wallet']));

    $fields['name'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Name or purpose of wallet'))
      ->setDescription(t('Other users need to know or guess this to transact with you.', [], ['context' => 'wallet name help']))
      ->setSetting('max_length', 128)
      ->setRequired(TRUE)
      ->setDisplayConfigurable('form', TRUE)
      ->setConstraints([
        'NotNull' => [],
        'WalletNameUnique' => []
      ]);
    $fields['created'] = BaseFieldDefinition::create('created')
      ->setLabel(t('Created on'))
      ->setDescription(t('The time that the wallet was created.'))
      ->setRevisionable(FALSE);
    // Following are computed fields, just for display esp by views
    $fields['balance'] = BaseFieldDefinition::create('worth')
      ->setLabel(t('Balance'))
      ->setDescription(t("Sum of all this wallet's credits minus debits"))
      ->setComputed(TRUE)
      ->setClass('\Drupal\mcapi\Plugin\Field\WalletBalance')
      ->setReadOnly(TRUE)
      ->setDisplayConfigurable('view', TRUE);
    $fields['volume'] = BaseFieldDefinition::create('worth')
      ->setLabel(t('Volume'))
      ->setDescription(t("Sum of all this wallet's transactions"))
      ->setComputed(TRUE)
      ->setClass('\Drupal\mcapi\Plugin\Field\WalletVolume')
      ->setReadOnly(TRUE)
      ->setDisplayConfigurable('view', TRUE);
    $fields['gross_in'] = BaseFieldDefinition::create('worth')
      ->setLabel(t('Gross income'))
      ->setDescription(t("Sum of all this wallet's income ever"))
      ->setComputed(TRUE)
      ->setClass('\Drupal\mcapi\Plugin\Field\WalletIncome')
      ->setReadOnly(TRUE)
      ->setDisplayConfigurable('view', TRUE);
    $fields['gross_out'] = BaseFieldDefinition::create('worth')
      ->setLabel(t('Gross expenditure'))
      ->setDescription(t("Sum of all this wallet's outgoings ever"))
      ->setComputed(TRUE)
      ->setClass('\Drupal\mcapi\Plugin\Field\WalletOutgoing')
      ->setReadOnly(TRUE)
      ->setDisplayConfigurable('view', TRUE);
    $fields['trades'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Trades'))
      ->setDescription(t("Number of transactions involving this wallet"))
      ->setComputed(TRUE)
      ->setClass('\Drupal\mcapi\Plugin\Field\WalletTrades')
      ->setReadOnly(TRUE)
      ->setDisplayConfigurable('view', TRUE);
    $fields['partners'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Number of partners'))
      ->setDescription(t("Number of unique trading partners"))
      ->setComputed(TRUE)
      ->setClass('\Drupal\mcapi\Plugin\Field\WalletPartners')
      ->setReadOnly(TRUE)
      ->setDisplayConfigurable('view', TRUE);

    return $fields;
  }

  private function loadStats() {
    $this->stats = $this->entityTypeManager()
      ->getStorage('mcapi_transaction')
      ->walletSummary($this->id());
  }

  /**
   * {@inheritDoc}
   */
  public function getAllStats($include_empty = FALSE) : array {
    if (!isset($this->stats)) {
      $this->loadStats();
    }
    if ($include_empty) {
      $result = $this->stats;
      foreach (Currency::loadMultiple() as $curr) {
        $result += [$curr->id => (object)self::STATS];
      }
      return $result;
    }
    return $this->stats;
  }

  /**
   * {@inheritDoc}
   */
  public function getStats(string $curr_id) : \stdClass {
    return $this->getAllStats(TRUE)[$curr_id];
  }

  /**
   * {@inheritDoc}
   */
  public function balance($curr_id, $display = CurrencyInterface::DISPLAY_NORMAL, $linked = TRUE) : string {
    return Currency::load($curr_id)
      ->format($this->getStats($curr_id)->balance, $display, $linked);
  }

  /**
   * {@inheritDoc}
   */
  public function getStatAll($stat = 'balance') : array {
    $worths = [];
    foreach ($this->getAllStats(TRUE) as $curr_id => $stats) {
      $worths[] = [
        'curr_id' => $curr_id,
        'value' => intval($stats->{$stat}),
      ];
    }
    return $worths;
  }

  /**
   * {@inheritDoc}
   */
  public function getCacheTagsToInvalidate() {
    $tags = parent::getCacheTagsToInvalidate();
    $tags[] = 'user:' . $this->getOwnerId();
    $tags[] = 'config:core.entity_view_display.mcapi_wallet';
    return $tags;
  }

  /**
   * {@inheritDoc}
   * Only virgin wallet can be deleted
   */
  public function isEmpty() : bool {
    $transactions = $this->entityTypeManager()
      ->getStorage('mcapi_transaction')->getQuery()->accessCheck(FALSE)
      ->involving($this)
      ->count()
      ->execute();
    return empty($transactions);
  }

}

==> mutual_credit-6.0.x-dev.tar.gz: mutual_credit/src/Entity/Wallet.php <==
<?php

namespace Drupal\mcapi\Entity;

use Drupal\user\EntityOwnerTrait;
use Drupal\user\EntityOwnerInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;

/**
 * Defines the wallet entity.
 *
 * The metaphor is not perfect because this wallet can hold a negaive balance.
 * A more accurate term, 'account', has another meaning in Drupal.
 *
 * This Entity knows all about the special intertrading wallets though none are
 * installed by default.
 *
 * @ContentEntityType(
 *   id = "mc_wallet",
 *   label = @Translation("Wallet"),
 *   module = "mcapi",
 *   handlers = {
 *     "view_builder" = "Drupal\mcapi\Entity\ViewBuilder\WalletViewBuilder",
 *     "storage" = "Drupal\mcapi\Entity\Storage\WalletStorage",
 *     "storage_schema" = "Drupal\mcapi\Entity\Storage\WalletStorageSchema",
 *     "list_builder" = "Drupal\mcapi\Entity\ListBuilder\WalletListBuilder",
 *     "access" = "Drupal\mcapi\Entity\Access\WalletAccessControlHandler",
 *     "form" = {
 *       "default" = "Drupal\mcapi\Form\WalletForm",
 *       "create" = "Drupal\mcapi\Form\WalletAddForm",
 *       "delete" = "Drupal\Core\Entity\ContentEntityDeleteForm",
 *     },
 *     "views_data" = "Drupal\mcapi\Entity\Views\WalletViewsData",
 *     "route_provider" = {
 *       "html" = "Drupal\Core\Entity\Routing\DefaultHtmlRouteProvider",
 *     },
 *   },
 *   admin_permission = "configure mcapi",
 *   base_table = "mc_wallet",
 *   entity_keys = {
 *     "label" = "name",
 *     "id" = "wid",
 *     "uuid" = "uuid",
 *     "owner" = "owner"
 *   },
 *   translatable = FALSE,
 *   links = {
 *     "canonical" = "/wallet/{mc_wallet}",
 *     "log" = "/wallet/{mc_wallet}/log",
 *     "inex" = "/wallet/{mc_wallet}/inex",
 *     "delete-form" = "/wallet/{mc_wallet}/delete",
 *     "edit-form" = "/wallet/{mc_wallet}/edit",
 *     "collection" = "/admin/accounting/wallets"
 *   },
 *   field_ui_base_route = "entity.mc_wallet.collection"
 * )
 */
class Wallet extends ContentEntityBase implements WalletInterface, EntityOwnerInterface {

  use MessengerTrait;
  use EntityOwnerTrait;

  const STATS = [
    'balance' => 0,
    'volume' => 0,
    'gross_in' => 0,
    'gross_out' => 0,
    'trades' => 0,
    'partners' => 0
  ];

  /**
   * {@inheritDoc}
   */
  public static function precreate($wallet, array &$values) {
    // NB owner is not set when generating a wallet for the field edit page.
    if (!isset($values['name']) and isset($values['owner'])) {
      // N.B. If this wallet is autocreated then the owner isn't saved yet
      $values['name'] = $values['owner']->label();
    }
  }

  /**
   * {@inheritDoc}
   */
  public function label() {
    if (\Drupal::currentUser()->isAnonymous()) {
      return t('Wallet @num', ['@num' => $this->id()]);
    }
    else {
      return $this->name->value;
    }
  }

  /**
   * {@inheritDoc}
   */
  public function doSave($id, EntityInterface $entity) {
    // Check the name length.
    if (strlen($this->name->value) > 128) {
      $this->name->value = substr($this->name->value, 0, 128);
      $this->messenger()->addWarning(
        t('Wallet name was truncated to 128 characters: @name', ['@name' => $this->name->value]),
        'warning'
      );
    }
  }

  /**
   * {@inheritDoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    $fields = parent::baseFieldDefinitions($entity_type)
      + static::ownerBaseFieldDefinitions($entity_type);
    $fields['owner']->setDisplayConfigurable('form', TRUE)
      ->setDisplayConfigurable('view', TRUE)
      ->setLabel(t('Owner user ID', [], ['context' => 'ID of the user who owns the wallet']));
    $fields['name'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Name or purpose of wallet'))
      ->setDescription(t('Other users need to know or guess this to transact with you.', [], ['context' => 'wallet name help']))
      ->setSetting('max_length', 128)
      ->setRequired(TRUE)
      ->setDisplayConfigurable('form', TRUE)
      ->setConstraints([
        'NotNull' => [],
        'WalletNameUnique' => []
      ]);
    $fields['created'] = BaseFieldDefinition::create('created')
      ->setLabel(t('Created on'))
      ->setDescription(t('The time that the wallet was created.'))
      ->setRevisionable(FALSE);
    return $fields;
  }

  /**
   * Magic method to get stats about the wallet.
   */
  public function &__get($name) {
    if (in_array($name, ['balance', 'volume', 'gross_in', 'gross_out', 'trades', 'partners'])) {
      // Returned by reference, these values can be edited but not saved.
      if (empty($this->values[$name])) {
        $stats = $this->entityTypeManager()
          ->getStorage('mc_wallet')
          ->getStats($this->id());
        $this->values = array_merge((array)$stats, $this->values);
      }
    }
    return parent::__get($name);
  }

  /**
   * {@inheritDoc}
   */
  public function getCacheTagsToInvalidate() {
    $tags = parent::getCacheTagsToInvalidate();
    $tags[] = 'config:core.entity_view_display.mc_wallet';
    return $tags;
  }

  /**
   * {@inheritDoc}
   * Only virgin wallet can be deleted
   */
  public function isEmpty() : bool {
    $transactions = $this->entityTypeManager()
      ->getStorage('mc_transaction')->getQuery()->accessCheck(FALSE)
      ->involving($this->id())
      ->count()
      ->execute();
    return empty($transactions);
  }

}
